/*	--*- c -*--
 * Copyright (C) 2016 Enrico Scholz <enrico.scholz@sigma-chemnitz.de>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; version 3 of the License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <stdio.h>
#include <stdbool.h>
#include "bayer2rgb.h"
#include "bayer2rgb-internal.h"

#define _always_inline_	__attribute__((__always_inline__))

static inline _always_inline_ unsigned int avg1(unsigned int max, int sft,
						unsigned int a)
{
	unsigned int		res = a;

	if (sft) {
		res += (1u << (sft-1));
		res >>= sft;
	}

	if (sft && res > max)
		res = max;

	return res;
}

static inline _always_inline_ unsigned int avg2(
	bool do_round, unsigned int max, int sft,
	unsigned int a, unsigned int b)
{
	unsigned int		res;

	if (!do_round)
		res = (a + b) / 2;
	else
		res = (a + b + 1) / 2;

	if (sft) {
		res += (1u << (sft-1));
		res >>= sft;
	}

	if ((sft || do_round) && res > max)
		res = max;

	return res;
}

static inline _always_inline_ unsigned int avg4(
	bool do_round, unsigned int max, int sft,
	unsigned int a, unsigned int b,
	unsigned int c, unsigned int d)
{
	unsigned int		res;

	if (!do_round)
		res = (a + b + c + d) / 4;
	else
		res = (a + b + c + d + 3) / 4;

	if (sft) {
		res += (1u << (sft-1));
		res >>= sft;
	}

	if ((sft || do_round) && res > max)
		res = max;

	return res;
}

#define _conv_gb(_p) {						\
		.r = avg2(round_2, max_rb, sft_rb,		\
			  (_p)[-stride], (_p)[+stride]),	\
		.b = avg2(round_2, max_rb, sft_rb,		\
			  (_p)[-1],      (_p)[+1]),		\
		.g = avg1(max_g, sft_g, (_p)[0]),		\
	}

#define _conv_bg(_p) {							\
		.r = avg4(round_4,  max_rb, sft_rb,			\
			  (_p)[-stride - 1], (_p)[-stride + 1],		\
			  (_p)[+stride - 1], (_p)[+stride + 1]),	\
		.b = avg1(max_rb, sft_rb, (_p)[0]),			\
		.g = avg4(round_4, max_g, sft_g,			\
			  (_p)[-1],      (_p)[+1],			\
			  (_p)[-stride], (_p)[+stride]),		\
	}

#define _conv_rg(_p) {							\
		.r = avg1(max_rb, sft_rb, (_p)[0]),			\
		.b = avg4(round_4, max_rb, sft_rb,			\
			  (_p)[-stride - 1], (_p)[-stride + 1],		\
			  (_p)[+stride - 1], (_p)[+stride + 1]),	\
		.g = avg4(round_4, max_g, sft_g,			\
			  (_p)[-stride], (_p)[+stride],			\
			  (_p)[-1],      (_p)[+1]),			\
	}

#define _conv_gr(_p) {						\
		.r = avg2(round_2, max_rb, sft_rb,		\
			  (_p)[-1], (_p)[+1]),			\
		.b = avg2(round_2, max_rb, sft_rb,		\
			  (_p)[-stride], (_p)[+stride]),	\
		.g = avg1(max_g, sft_g, (_p)[0]),		\
	}

void bayer2rgb_convert_c(struct image_in const *input,
			 struct image_out const *output,
			 struct image_conversion_info *info)
{
	static unsigned int const	border = 2;
	int				handled = -1;

#define CONVERT_BODY	"convert-c.inc.h"

	switch (input->type) {
	case BAYER_GBRG:
#define conv_fn_00	_conv_gb
#define conv_fn_01	_conv_bg
#define conv_fn_10	_conv_rg
#define conv_fn_11	_conv_gr

#include "convert-c-body-outer.inc.h"
		break;

	case BAYER_RGGB:
#define conv_fn_00	_conv_rg
#define conv_fn_01	_conv_gr
#define conv_fn_10	_conv_gb
#define conv_fn_11	_conv_bg

#include "convert-c-body-outer.inc.h"
		break;

	case BAYER_GRBG:
#define conv_fn_00	_conv_gr
#define conv_fn_01	_conv_rg
#define conv_fn_10	_conv_bg
#define conv_fn_11	_conv_gb

#include "convert-c-body-outer.inc.h"
		break;

	case BAYER_BGGR:
#define conv_fn_00	_conv_bg
#define conv_fn_01	_conv_gb
#define conv_fn_10	_conv_gr
#define conv_fn_11	_conv_rg

#include "convert-c-body-outer.inc.h"
		break;

	default:
		abort();
	}

	if (!handled)
		abort();
}
